# -*- coding: utf-8 -*-
from collective.easyform.tests import base
from os import listdir
from os import path
from plone.testing import layered
from robotsuite import RobotTestSuite
from unittest import TestSuite


def test_suite():
    suite = TestSuite()
    current_dir = path.abspath(path.dirname(__file__))
    robot_dir = path.join(current_dir, 'robot')
    robot_tests = [
        path.join('robot', doc)
        for doc in listdir(robot_dir)
        if doc.startswith('test') and doc.endswith('.robot')
    ]
    for test in robot_tests:
        suite.addTests([
            layered(
                RobotTestSuite(test),
                layer=base.ROBOT_TESTING
            ),
        ])
    suite.level = 3
    return suite



# -*- coding: utf-8 -*-
from .validators import isTALES
from collective.easyform import easyformMessageFactory as _  # NOQA
from collective.easyform import config
from plone.autoform import directives
from plone.schemaeditor.interfaces import ID_RE
from plone.schemaeditor.interfaces import IFieldContext
from plone.schemaeditor.interfaces import IFieldEditorExtender
from plone.schemaeditor.interfaces import ISchemaContext
from plone.supermodel.model import fieldset
from plone.supermodel.model import Schema
from zope.interface import Interface

import z3c.form.interfaces
import zope.interface
import zope.schema.interfaces


try:
    from plone.schemaeditor import SchemaEditorMessageFactory as __  # NOQA
except ImportError:
    from plone.schemaeditor import _ as __


MODIFY_PORTAL_CONTENT = 'cmf.ModifyPortalContent'


class IEasyFormActionsEditorExtender(IFieldEditorExtender):
    pass


def isValidFieldName(value):
    if not ID_RE.match(value):
        raise zope.interface.Invalid(
            __(u'Please use only letters, numbers and '
               u'the following characters: _.'))
    return True


class INewAction(Interface):

    title = zope.schema.TextLine(
        title=__(u'Title'),
        required=True
    )

    __name__ = zope.schema.ASCIILine(
        title=__(u'Short Name'),
        description=__(u'Used for programmatic access to the field.'),
        required=True,
        constraint=isValidFieldName,
    )

    description = zope.schema.Text(
        title=__(u'Help Text'),
        description=__(u'Shows up in the form as help text for the field.'),
        required=False
    )

    factory = zope.schema.Choice(
        title=_(u'Action type'),
        vocabulary='EasyFormActions',
        required=True,
    )

    @zope.interface.invariant
    def checkTitleAndDescriptionTypes(data):  # NOQA
        if data.__name__ is not None and data.factory is not None:
            if (
                data.__name__ == 'title' and
                data.factory.fieldcls is not zope.schema.TextLine
            ):
                raise zope.interface.Invalid(
                    __(
                        u"The 'title' field must be a Text line (string) "
                        u"field."
                    )
                )
            if (
                data.__name__ == 'description' and
                data.factory.fieldcls is not zope.schema.Text
            ):
                raise zope.interface.Invalid(
                    __(u"The 'description' field must be a Text field."))


class IActionFactory(zope.schema.interfaces.IField):

    """ A component that instantiates a action when called.
    """
    title = zope.schema.TextLine(title=__(u'Title'))


class IEasyFormActionsContext(ISchemaContext):

    """
    EasyForm actions view interface
    """


class IActionExtender(Schema):
    fieldset(u'overrides', label=_('Overrides'), fields=['execCondition'])
    directives.read_permission(execCondition=MODIFY_PORTAL_CONTENT)
    directives.write_permission(execCondition=config.EDIT_TALES_PERMISSION)
    execCondition = zope.schema.TextLine(
        title=_(u'label_execcondition_text', default=u'Execution Condition'),
        description=_(
            u'help_execcondition_text',
            default=u'A TALES expression that will be evaluated to determine '
                    u'whether or not to execute this action. Leave empty if '
                    u'unneeded, and the action will be executed. Your '
                    u'expression should evaluate as a boolean; return True '
                    u'if you wish the action to execute. PLEASE NOTE: errors '
                    u'in the evaluation of this expression will  cause an '
                    u'error on form display.'
        ),
        default=u'',
        constraint=isTALES,
        required=False,
    )


class IEasyFormActionContext(IFieldContext):

    """
    EasyForm action content marker
    """


class IActionEditForm(z3c.form.interfaces.IEditForm):

    """ Marker interface for action edit forms
    """


class IAction(Schema, zope.schema.interfaces.IField):
    directives.omitted('required', 'order', 'default',
                       'missing_value', 'readonly')
#     required = zope.schema.Bool(
#         title=_('Enabled'),
#        description=_('Tells whether a action is enabled.'),
#         default=True)

    def onSuccess(fields, request):  # NOQA
        pass



# -*- coding: utf-8 -*-
#
# Test EasyForm initialisation and set-up
#

from collective.easyform.tests import base
from Products.CMFCore.utils import getToolByName

import Products


def getAddPermission(product, name):
    ''' find the add permission for a meta_type '''

    name = '{0}: {1}'.format(product, name)
    for mt in Products.meta_types:
        if mt['name'] == name:
            return mt['permission']
    return ''


class TestInstallation(base.EasyFormTestCase):

    '''Ensure product is properly installed'''

    def afterSetUp(self):
        base.EasyFormTestCase.afterSetUp(self)

        self.types = self.portal.portal_types
        self.properties = self.portal.portal_properties
        self.controlpanel = self.portal.portal_controlpanel

        self.metaTypes = ('EasyForm',)

    def testTypesInstalled(self):
        for t in self.metaTypes:
            self.assertTrue(t in self.types.objectIds())

    def testTypeActions(self):
        # hide properties/references tabs
        for typ in self.metaTypes:
            for act in self.types[typ].listActions():
                if act.id in ['metadata', 'references']:
                    self.assertFalse(act.visible)
                if act.id in ['actions', 'fields']:
                    self.assertTrue(act.visible)

    def testArchetypesToolCatalogRegistration(self):
        at_tool = getToolByName(self.portal, 'archetype_tool', None)
        if not at_tool:
            return
        for t in self.metaTypes:
            self.assertEqual(1, len(at_tool.getCatalogsByType(t)))
            self.assertEqual(
                'portal_catalog', at_tool.getCatalogsByType(t)[0].getId())

    def ttestControlPanelConfigletInstalled(self):
        self.assertTrue(
            'EasyForm' in [
                action.id for action in self.controlpanel.listActions()
            ]
        )

    def ttestAddPermissions(self):
        ''' Test to make sure add permissions are as intended '''

        ADD_CONTENT_PERMISSION = 'EasyForm: Add Content'
        CSA_ADD_CONTENT_PERMISSION = 'EasyForm: Add Custom Scripts'
        MA_ADD_CONTENT_PERMISSION = 'EasyForm: Add Mailers'
        SDA_ADD_CONTENT_PERMISSION = 'EasyForm: Add Data Savers'

        self.assertEqual(
            getAddPermission('EasyForm', 'EasyForm'),
            ADD_CONTENT_PERMISSION
        )
        self.assertEqual(
            getAddPermission('EasyForm', 'Mailer Adapter'),
            MA_ADD_CONTENT_PERMISSION
        )
        self.assertEqual(
            getAddPermission('EasyForm', 'Save Data Adapter'),
            SDA_ADD_CONTENT_PERMISSION
        )
        self.assertEqual(
            getAddPermission('EasyForm', 'Custom Script Adapter'),
            CSA_ADD_CONTENT_PERMISSION
        )

    def testActionsInstalled(self):
        self.assertTrue(
            self.portal.portal_actions.getActionInfo('object_buttons/export')
        )
        self.assertTrue(
            self.portal.portal_actions.getActionInfo('object_buttons/import')
        )
        self.assertTrue(
            self.portal.portal_actions.getActionInfo(
                'object_buttons/saveddata'
            )
        )

    def testModificationsToPropSheetLinesNotPuged(self):
        property_mappings = [{
            'propsheet': 'site_properties',
            'added_props': [
                'use_folder_tabs',
                'typesLinkToFolderContentsInFC',
                'default_page_types',
            ]
        }]

        # add garbage prop element to each lines property
        for mapping in property_mappings:
            sheet = self.properties[mapping['propsheet']]
            for lines_prop in mapping['added_props']:
                propitems = list(sheet.getProperty(lines_prop))
                propitems.append('foo')
                sheet.manage_changeProperties({lines_prop: propitems})

        # reinstall
        qi = self.portal.portal_quickinstaller
        qi.reinstallProducts(['collective.easyform'])

        # now make sure our garbage values survived the reinstall
        for mapping in property_mappings:
            sheet = self.properties[mapping['propsheet']]
            for lines_prop in mapping['added_props']:
                self.assertTrue(
                    'foo' in sheet.getProperty(lines_prop),
                    "Our garbage item didn't survive reinstall for property "
                    "{0} within property sheet {1}".format(
                        lines_prop,
                        mapping['propsheet']
                    )
                )

    def test_EasyFormInDefaultPageTypes(self):
        propsTool = getToolByName(self.portal, 'portal_properties')
        siteProperties = getattr(propsTool, 'site_properties')
        defaultPageTypes = list(
            siteProperties.getProperty('default_page_types'))
        self.assertTrue('EasyForm' in defaultPageTypes)

    def testTypeViews(self):
        self.assertEqual(
            self.types.EasyForm.getAvailableViewMethods(self.types), ('view',))



# -*- coding: utf-8 -*-
from AccessControl import getSecurityManager
from BTrees.IOBTree import IOBTree
from BTrees.LOBTree import LOBTree as SavedDataBTree
from collective.easyform import easyformMessageFactory as _
from collective.easyform.api import dollar_replacer
from collective.easyform.api import filter_fields
from collective.easyform.api import filter_widgets
from collective.easyform.api import format_addresses
from collective.easyform.api import get_context
from collective.easyform.api import get_expression
from collective.easyform.api import get_schema
from collective.easyform.api import is_file_data
from collective.easyform.api import lnbr
from collective.easyform.api import OrderedDict
from collective.easyform.interfaces import IAction
from collective.easyform.interfaces import IActionFactory
from collective.easyform.interfaces import ICustomScript
from collective.easyform.interfaces import IExtraData
from collective.easyform.interfaces import IMailer
from collective.easyform.interfaces import ISaveData
from copy import deepcopy
from csv import writer as csvwriter
from DateTime import DateTime
from email import Encoders
from email.Header import Header
from email.MIMEAudio import MIMEAudio
from email.MIMEBase import MIMEBase
from email.MIMEImage import MIMEImage
from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from email.utils import formataddr
from io import BytesIO
from json import dumps
from logging import getLogger
from plone import api
from plone.app.textfield.value import RichTextValue
from plone.autoform.view import WidgetsView
from plone.supermodel.exportimport import BaseHandler
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.PythonScripts.PythonScript import PythonScript
from StringIO import StringIO
from time import time
from xml.etree import ElementTree as ET
from z3c.form.interfaces import DISPLAY_MODE
from zope.component import queryUtility
from zope.contenttype import guess_content_type
from zope.interface import implementer
from zope.schema import Bool
from zope.schema import getFieldsInOrder
from zope.security.interfaces import IPermission


logger = getLogger('collective.easyform')


@implementer(IActionFactory)
class ActionFactory(object):

    title = u''

    def __init__(self, fieldcls, title, permission, *args, **kw):
        self.fieldcls = fieldcls
        self.title = title
        self.permission = permission
        self.args = args
        self.kw = kw

    def available(self, context):
        """ field is addable in the current context """
        securityManager = getSecurityManager()
        permission = queryUtility(IPermission, name=self.permission)
        if permission is None:
            return True
        return bool(securityManager.checkPermission(permission.title, context))

    def editable(self, field):
        """ test whether a given instance of a field is editable """
        return True

    def __call__(self, *args, **kw):
        kwargs = deepcopy(self.kw)
        kwargs.update(**kw)
        return self.fieldcls(*(self.args + args), **kwargs)


@implementer(IAction)
class Action(Bool):
    """ Base action class """

    def onSuccess(self, fields, request):
        raise NotImplementedError(
            "There is not implemented 'onSuccess' of {0!r}".format(self))


class DummyFormView(WidgetsView):
    """ A dummy form to get the widgets rendered for the mailer action
    """

    mode = DISPLAY_MODE
    ignoreContext = True
    ignoreRequest = False


@implementer(IMailer)
class Mailer(Action):
    __doc__ = IMailer.__doc__

    def __init__(self, **kw):
        for i, f in IMailer.namesAndDescriptions():
            setattr(self, i, kw.pop(i, f.default))
        super(Mailer, self).__init__(**kw)

    def get_portal_email_address(self, context):
        """Return the email address defined in the Plone site."""
        return api.portal.get_registry_record('plone.email_from_address')

    def secure_header_line(self, line):
        if not line:
            return ''
        nlpos = line.find('\x0a')
        if nlpos >= 0:
            line = line[:nlpos]
        nlpos = line.find('\x0d')
        if nlpos >= 0:
            line = line[:nlpos]
        return line

    def get_mail_body(self, unsorted_data, request, context):
        """Returns the mail-body with footer.
        """
        schema = get_schema(context)

        form = DummyFormView(context, request)
        form.schema = schema
        form.prefix = 'form'
        form._update()
        widgets = filter_widgets(self, form.w)
        data = filter_fields(self, schema, unsorted_data)

        bodyfield = self.body_pt

        # pass both the bare_fields (fgFields only) and full fields.
        # bare_fields for compatability with older templates,
        # full fields to enable access to htmlValue
        if isinstance(self.body_pre, basestring):
            body_pre = self.body_pre
        else:
            body_pre = self.body_pre.output

        if isinstance(self.body_post, basestring):
            body_post = self.body_post
        else:
            body_post = self.body_post.output

        if isinstance(self.body_footer, basestring):
            body_footer = self.body_footer
        else:
            body_footer = self.body_footer.output

        extra = {
            'data': data,
            'fields': OrderedDict([
                (i, j.title)
                for i, j in getFieldsInOrder(schema)
            ]),
            'widgets': widgets,
            'mailer': self,
            'body_pre': body_pre and lnbr(dollar_replacer(body_pre, data)),
            'body_post': body_post and lnbr(dollar_replacer(body_post, data)),
            'body_footer': body_footer and lnbr(
                dollar_replacer(body_footer, data)),
        }
        template = ZopePageTemplate(self.__name__)
        template.write(bodyfield)
        template = template.__of__(context)
        return template.pt_render(extra_context=extra)

    def get_owner_info(self, context):
        """Return owner info
        """
        pms = getToolByName(context, 'portal_membership')
        ownerinfo = context.getOwner()
        ownerid = fullname = ownerinfo.getId()
        userdest = pms.getMemberById(ownerid)
        if userdest is not None:
            fullname = userdest.getProperty('fullname', ownerid)
        toemail = ''
        if userdest is not None:
            toemail = userdest.getProperty('email', '')
        if not toemail:
            toemail = self.get_portal_email_address(context)
        if not toemail:
            raise ValueError(
                u"Unable to mail form input because no recipient address has "
                u"been specified. Please check the recipient settings of the "
                u"EasyForm Mailer within the current form folder."
            )
        return (fullname, toemail)

    def get_addresses(
        self,
        fields,
        request,
        context,
        from_addr=None,
        to_addr=None
    ):
        """
        Return addresses
        """
        # get Reply-To
        reply_addr = None
        if hasattr(self, 'replyto_field'):
            reply_addr = fields.get(self.replyto_field, None)

        # Get From address
        portal_addr = self.get_portal_email_address(context)
        from_addr = from_addr or portal_addr

        if hasattr(self, 'senderOverride') and self.senderOverride:
            _from = get_expression(context, self.senderOverride, fields=fields)
            if _from:
                from_addr = _from

        # Get To address and full name
        recip_email = None
        if hasattr(self, 'to_field') and self.to_field:
            recip_email = fields.get(self.to_field, None)
        if not recip_email:
            if self.recipient_email != '':
                recip_email = self.recipient_email

        if hasattr(self, 'recipientOverride') and self.recipientOverride:
            _recip = get_expression(
                context,
                self.recipientOverride,
                fields=fields
            )
            if _recip:
                recip_email = _recip

        to = None
        if to_addr:
            to = format_addresses(to_addr)
        elif recip_email:
            to = format_addresses(
                recip_email,
                self.recipient_name
            )
        else:
            # Use owner adress or fall back to portal email_from_address.
            to = formataddr(self.get_owner_info(context))

        assert(to)
        return (to, from_addr, reply_addr)

    def get_subject(self, fields, request, context):
        """Return subject
        """
        # get subject header
        nosubject = u'(no subject)'  # TODO: translate
        subject = None
        if hasattr(self, 'subjectOverride') and self.subjectOverride:
            # subject has a TALES override
            subject = get_expression(
                context,
                self.subjectOverride,
                fields=fields
            ).strip()

        if not subject:
            subject = getattr(self, 'msg_subject', nosubject)
            subjectField = fields.get(self.subject_field, None)
            if subjectField is not None:
                subject = subjectField
            else:
                # we only do subject expansion if there's no field chosen
                subject = dollar_replacer(subject, fields)

        if isinstance(subject, basestring):
            subject = safe_unicode(subject)
        elif subject and isinstance(subject, (set, tuple, list)):
            subject = ', '.join([safe_unicode(s) for s in subject])
        else:
            subject = nosubject

        # transform subject into mail header encoded string
        email_charset = 'utf-8'
        msgSubject = self.secure_header_line(
            subject).encode(email_charset, 'replace')
        msgSubject = str(Header(msgSubject, email_charset))
        return msgSubject

    def get_header_info(self, fields, request, context,
                        from_addr=None, to_addr=None,
                        subject=None):
        """Return header info

        header info is a dictionary

        Keyword arguments:
        request -- (optional) alternate request object to use
        """
        (to, from_addr, reply) = self.get_addresses(fields, request, context)

        headerinfo = OrderedDict()
        headerinfo['To'] = self.secure_header_line(to)
        headerinfo['From'] = self.secure_header_line(from_addr)
        if reply:
            headerinfo['Reply-To'] = self.secure_header_line(reply)
        headerinfo['Subject'] = self.get_subject(fields, request, context)

        # CC
        cc_recips = filter(None, self.cc_recipients)
        if hasattr(self, 'ccOverride') and self.ccOverride:
            _cc = get_expression(context, self.ccOverride, fields=fields)
            if _cc:
                cc_recips = _cc

        if cc_recips:
            headerinfo['Cc'] = format_addresses(cc_recips)

        # BCC
        bcc_recips = filter(None, self.bcc_recipients)
        if hasattr(self, 'bccOverride') and self.bccOverride:
            _bcc = get_expression(context, self.bccOverride, fields=fields)
            if _bcc:
                bcc_recips = _bcc

        if bcc_recips:
            headerinfo['Bcc'] = format_addresses(bcc_recips)

        for key in getattr(self, 'xinfo_headers', []):
            headerinfo['X-{0}'.format(key)] = self.secure_header_line(
                request.get(key, 'MISSING'))

        return headerinfo

    def serialize(self, field):
        """Serializa field to save to XML.
        """
        if isinstance(field, set) or isinstance(field, list):
            list_value = list([str(f) for f in field])
            return dumps(list_value)
        if isinstance(field, RichTextValue):
            return field.raw
        return str(field)

    def get_attachments(self, fields, request):
        """Return all attachments uploaded in form.
        """

        attachments = []

        # if requested, generate CSV attachment of form values
        sendCSV = getattr(self, 'sendCSV', None)
        if sendCSV:
            csvdata = []
        sendXML = getattr(self, 'sendXML', None)
        if sendXML:
            xmlRoot = ET.Element("form")
        for fname in fields:
            field = fields[fname]
            showFields = getattr(self, 'showFields', []) or []

            if sendCSV:
                if not is_file_data(field) and (
                        getattr(self, 'showAll', True) or fname in showFields):
                    csvdata.append(field)

            if sendXML:
                if not is_file_data(field) and (
                        getattr(self, 'showAll', True) or fname in showFields):
                    ET.SubElement(xmlRoot, "field", name=fname).text \
                        = self.serialize(field)

            if is_file_data(field) and (
                    getattr(self, 'showAll', True) or fname in showFields):
                data = field.data
                filename = field.filename
                mimetype, enc = guess_content_type(filename, data, None)
                attachments.append((filename, mimetype, enc, data))

        if sendCSV:
            output = BytesIO()
            writer = csvwriter(output)
            writer.writerow(csvdata)
            csv = output.getvalue()
            now = DateTime().ISO().replace(' ', '-').replace(':', '')
            filename = 'formdata_{0}.csv'.format(now)
            attachments.append((filename, 'text/plain', 'utf-8', csv))

        if sendXML:
            xmlstr = ET.tostring(xmlRoot, encoding='utf8', method='xml')
            now = DateTime().ISO().replace(' ', '-').replace(':', '')
            filename = 'formdata_{0}.xml'.format(now)
            attachments.append((filename, 'text/xml', 'utf-8', xmlstr))

        return attachments

    def get_mail_text(self, fields, request, context):
        """Get header and body of e-mail as text (string)
        """
        headerinfo = self.get_header_info(fields, request, context)
        body = self.get_mail_body(fields, request, context)
        if not isinstance(body, unicode):
            body = unicode(body, 'UTF-8')
        email_charset = 'utf-8'
        # always use text/plain for encrypted bodies
        subtype = getattr(
            self, 'gpg_keyid', False) and 'plain' or self.body_type or 'html'
        mime_text = MIMEText(body.encode(email_charset, 'replace'),
                             _subtype=subtype, _charset=email_charset)

        attachments = self.get_attachments(fields, request)

        if attachments:
            outer = MIMEMultipart()
            outer.attach(mime_text)
        else:
            outer = mime_text

        # write header
        for key, value in headerinfo.items():
            outer[key] = value

        # write additional header
        additional_headers = self.additional_headers or []
        for a in additional_headers:
            key, value = a.split(':', 1)
            outer.add_header(key, value.strip())

        for attachment in attachments:
            filename = attachment[0]
            ctype = attachment[1]
            # encoding = attachment[2]
            content = attachment[3]

            if ctype is None:
                ctype = 'application/octet-stream'

            maintype, subtype = ctype.split('/', 1)

            if maintype == 'text':
                msg = MIMEText(content, _subtype=subtype)
            elif maintype == 'image':
                msg = MIMEImage(content, _subtype=subtype)
            elif maintype == 'audio':
                msg = MIMEAudio(content, _subtype=subtype)
            else:
                msg = MIMEBase(maintype, subtype)
                msg.set_payload(content)
                # Encode the payload using Base64
                Encoders.encode_base64(msg)

            # Set the filename parameter
            if isinstance(filename, unicode):
                filename = filename.encode('utf-8')
            msg.add_header(
                'Content-Disposition', 'attachment',
                filename=('utf-8', '', filename))
            outer.attach(msg)

        return outer.as_string()

    def onSuccess(self, fields, request):
        """
        e-mails data.
        """
        context = get_context(self)
        mailtext = self.get_mail_text(fields, request, context)
        host = api.portal.get_tool(name='MailHost')
        host.send(mailtext)


@implementer(ICustomScript)
class CustomScript(Action):
    __doc__ = ICustomScript.__doc__

    def __init__(self, **kw):
        for i, f in ICustomScript.namesAndDescriptions():
            setattr(self, i, kw.pop(i, f.default))
        super(CustomScript, self).__init__(**kw)

    def getScript(self, context):
        # Generate Python script object

        body = self.ScriptBody
        role = self.ProxyRole
        script = PythonScript(self.__name__)
        script = script.__of__(context)

        # Skip check roles
        script._validateProxy = lambda i=None: None

        # Force proxy role
        if role != u'none':
            script.manage_proxy((role,))

        body = body.encode('utf-8')
        params = 'fields, easyform, request'
        script.ZPythonScript_edit(params, body)
        return script

    def sanifyFields(self, form):
        # Makes request.form fields accessible in a script
        #
        # Avoid Unauthorized exceptions since request.form is inaccesible

        result = {}
        for field in form:
            result[field] = form[field]
        return result

    def checkWarningsAndErrors(self, script):
        # Raise exception if there has been bad things with the script
        # compiling

        if len(script.warnings) > 0:
            logger.warn('Python script ' + self.__name__ +
                        ' has warning:' + str(script.warnings))

        if len(script.errors) > 0:
            logger.error('Python script ' + self.__name__ +
                         ' has errors: ' + str(script.errors))
            raise ValueError(
                'Python script {0} has errors: {1}'.format(
                    self.__name__, str(script.errors)
                )
            )

    def executeCustomScript(self, result, form, req):
        # Execute in-place script

        # @param result Extracted fields from request.form
        # @param form EasyForm object

        script = self.getScript(form)
        self.checkWarningsAndErrors(script)
        response = script(result, form, req)
        return response

    def onSuccess(self, fields, request):
        # Executes the custom script
        form = get_context(self)
        resultData = self.sanifyFields(request.form)
        return self.executeCustomScript(resultData, form, request)


@implementer(ISaveData)
class SaveData(Action):
    __doc__ = ISaveData.__doc__

    def __init__(self, **kw):
        for i, f in ISaveData.namesAndDescriptions():
            setattr(self, i, kw.pop(i, f.default))
        super(SaveData, self).__init__(**kw)

    @property
    def _storage(self):
        context = get_context(self)
        if not hasattr(context, '_inputStorage'):
            context._inputStorage = {}
        if self.__name__ not in context._inputStorage:
            context._inputStorage[self.__name__] = SavedDataBTree()
        return context._inputStorage[self.__name__]

    def clearSavedFormInput(self):
        # convenience method to clear input buffer
        self._storage.clear()

    def getSavedFormInput(self):
        """ returns saved input as an iterable;
            each row is a sequence of fields.
        """

        return self._storage.values()

    def getSavedFormInputItems(self):
        """ returns saved input as an iterable;
            each row is an (id, sequence of fields) tuple
        """
        return self._storage.items()

    def getSavedFormInputForEdit(self, header=False, delimiter=','):
        """ returns saved as CSV text """
        sbuf = StringIO()
        writer = csvwriter(sbuf, delimiter=delimiter)
        names = self.getColumnNames()
        titles = self.getColumnTitles()

        if header:
            encoded_titles = []
            for t in titles:
                if isinstance(t, unicode):
                    t = t.encode('utf-8')
                encoded_titles.append(t)
            writer.writerow(encoded_titles)
        for row in self.getSavedFormInput():
            def get_data(row, i):
                data = row.get(i, '')
                if is_file_data(data):
                    data = data.filename
                if isinstance(data, unicode):
                    return data.encode('utf-8')
                return data
            writer.writerow([get_data(row, i) for i in names])
        res = sbuf.getvalue()
        sbuf.close()
        return res

    def getColumnNames(self):
        # """Returns a list of column names"""
        context = get_context(self)
        showFields = getattr(self, 'showFields', [])
        if showFields is None:
            showFields = []
        names = [
            name
            for name, field in getFieldsInOrder(get_schema(context))
            if not showFields or name in showFields
        ]
        if self.ExtraData:
            for f in self.ExtraData:
                names.append(f)
        return names

    def getColumnTitles(self):
        # """Returns a list of column titles"""
        context = get_context(self)
        showFields = getattr(self, 'showFields', [])
        if showFields is None:
            showFields = []

        names = [
            field.title
            for name, field in getFieldsInOrder(get_schema(context))
            if not showFields or name in showFields
        ]
        if self.ExtraData:
            for f in self.ExtraData:
                names.append(IExtraData[f].title)
        return names

    def download_csv(self, response):
        # """Download the saved data as csv
        # """
        response.setHeader(
            'Content-Disposition',
            'attachment; filename="{0}.csv"'.format(self.__name__)
        )
        response.setHeader('Content-Type', 'text/comma-separated-values')
        response.write(self.getSavedFormInputForEdit(
            getattr(self, 'UseColumnNames', False), delimiter=','))

    def download_tsv(self, response):
        # """Download the saved data as tsv
        # """
        response.setHeader(
            'Content-Disposition',
            'attachment; filename="{0}.tsv"'.format(self.__name__)
        )
        response.setHeader('Content-Type', 'text/tab-separated-values')
        response.write(self.getSavedFormInputForEdit(
            getattr(self, 'UseColumnNames', False), delimiter='\t'))

    def download(self, response):
        # """Download the saved data
        # """
        format = getattr(self, 'DownloadFormat', 'tsv')
        if format == 'tsv':
            return self.download_tsv(response)
        else:
            assert format == 'csv', 'Unknown download format'
            return self.download_csv(response)

    def itemsSaved(self):
        return len(self._storage)

    def delDataRow(self, key):
        del self._storage[key]

    def setDataRow(self, key, value):
        # sdata = self.storage[id]
        # sdata.update(data)
        # self.storage[id] = sdata
        self._storage[key] = value

    def addDataRow(self, value):
        storage = self._storage
        if isinstance(storage, IOBTree):
            # 32-bit IOBTree; use a key which is more likely to conflict
            # but which won't overflow the key's bits
            id = storage.maxKey() + 1
        else:
            # 64-bit LOBTree
            id = int(time() * 1000)
            while id in storage:  # avoid collisions during testing
                id += 1
        value['id'] = id
        storage[id] = value

    def onSuccess(self, fields, request):
        """
        saves data.
        """
        # if LP_SAVE_TO_CANONICAL and not loopstop:
        # LinguaPlone functionality:
        # check to see if we're in a translated
        # form folder, but not the canonical version.
        # parent = self.aq_parent
        # if safe_hasattr(parent, 'isTranslation') and \
        # parent.isTranslation() and not parent.isCanonical():
        # look in the canonical version to see if there is
        # a matching (by id) save-data adapter.
        # If so, call its onSuccess method
        # cf = parent.getCanonical()
        # target = cf.get(self.getId())
        # if target is not None and target.meta_type == 'FormSaveDataAdapter':
        # target.onSuccess(fields, request, loopstop=True)
        # return
        data = {}
        showFields = getattr(self, 'showFields', []) or self.getColumnNames()
        for f in fields:
            if f not in showFields:
                continue
            data[f] = fields[f]

        if self.ExtraData:
            for f in self.ExtraData:
                if f == 'dt':
                    data[f] = str(DateTime())
                else:
                    data[f] = getattr(request, f, '')

        self.addDataRow(data)


MailerAction = ActionFactory(
    Mailer, _
    (u'label_mailer_action', default=u'Mailer'),
    'collective.easyform.AddMailers'
)
CustomScriptAction = ActionFactory(
    CustomScript, _
    (u'label_customscript_action', default=u'Custom Script'),
    'collective.easyform.AddCustomScripts'
)
SaveDataAction = ActionFactory(
    SaveData, _
    (u'label_savedata_action', default=u'Save Data'),
    'collective.easyform.AddDataSavers'
)

MailerHandler = BaseHandler(Mailer)
CustomScriptHandler = BaseHandler(CustomScript)
SaveDataHandler = BaseHandler(SaveData)



# -*- coding: utf-8 -*-
#
# Integeration tests specific to the mailer
#

from collective.easyform.actions import DummyFormView
from collective.easyform.api import filter_widgets
from collective.easyform.api import get_schema
from collective.easyform.tests import base


class TestFunctions(base.EasyFormTestCase):
    """ Test mailer action """

    def afterSetUp(self):
        super(TestFunctions, self).afterSetUp()
        self.folder.invokeFactory('EasyForm', 'ff1')
        self.ff1 = getattr(self.folder, 'ff1')
        self.dummy_form = DummyFormView(self.ff1, self.layer['request'])
        self.dummy_form.schema = get_schema(self.ff1)
        self.dummy_form.prefix = 'form'
        self.dummy_form._update()

    def test_selective_widgets(self):
        """ Test selective inclusion of widgets for mail and thank you page """

        self.assertEqual(
            filter_widgets(self.ff1, self.dummy_form.w).keys(),
            ['replyto', 'topic', 'comments']
        )

        self.ff1.showFields = ('topic', 'comments',)
        self.assertEqual(
            filter_widgets(self.ff1, self.dummy_form.w).keys(),
            ['replyto', 'topic', 'comments']
        )

        self.ff1.showAll = False
        self.assertEqual(
            filter_widgets(self.ff1, self.dummy_form.w).keys(),
            ['topic', 'comments']
        )



from setuptools import setup, find_packages

version = '2.0.0'

setup(
    name='collective.easyform',
    version=version,
    description='Forms for Plone',
    long_description=open('README.rst').read() +
    '\n' + open('CHANGES.rst').read(),
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Plone',
        'Framework :: Plone :: 5.1',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    ],
    keywords='form plone ttw builder email',
    author='Roman Kozlovskyi',
    author_email='krzroman@gmail.com',
    url='https://github.com/collective/collective.easyform',
    license='GPLv2',
    packages=find_packages('src', exclude=['ez_setup']),
    namespace_packages=['collective'],
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'plone.api',
        'plone.app.dexterity',
        'plone.app.textfield >= 1.2.8',
        'plone.autoform',
        'plone.dexterity',
        'plone.namedfile',
        'plone.schema',
        'plone.schemaeditor >= 2.0.18',
        'plone.supermodel',
        'Products.CMFPlone',
        'Products.validation',
        # -*- Extra requirements: -*-
    ],
    extras_require={
        'recaptcha': [
            'plone.formwidget.recaptcha'
        ],
        'test': [
            'plone.app.testing[robot]',
            'plone.app.robotframework',
            'plone.app.contenttypes',
        ]
    },
    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



# -*- coding: utf-8 -*-
from collective.easyform import easyformMessageFactory as _
from collective.easyform.api import get_expression
from collective.easyform.interfaces import IEasyForm
from collective.easyform.interfaces import IEasyFormForm
from collective.easyform.interfaces import IFieldExtender
from collective.easyform.interfaces import ILabel
from collective.easyform.interfaces import IReCaptcha
from collective.easyform.interfaces import IRichLabel
from collective.easyform.validators import IFieldValidator
from plone.schemaeditor.fields import FieldFactory
from plone.supermodel.exportimport import BaseHandler
from z3c.form import validator as z3c_validator
from z3c.form.interfaces import IValidator
from z3c.form.interfaces import IValue
from zope.component import adapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import Invalid
from zope.schema import Field
from zope.schema import TextLine
from zope.schema._bootstrapinterfaces import IFromUnicode
from zope.schema.interfaces import IField


@implementer(IValidator)
@adapter(IEasyForm, Interface, IEasyFormForm, IField, Interface)
class FieldExtenderValidator(z3c_validator.SimpleFieldValidator):

    """ z3c.form validator class for easyform fields """

    def validate(self, value):
        """ Validate field by TValidator """
        super(FieldExtenderValidator, self).validate(value)
        efield = IFieldExtender(self.field)
        validators = getattr(efield, 'validators', [])
        if validators:
            for validator in validators:
                vmethod = queryUtility(IFieldValidator, name=validator)
                if not vmethod:
                    continue
                res = vmethod(value)
                if res:
                    raise Invalid(res)
        TValidator = getattr(efield, 'TValidator', None)
        if TValidator:
            try:
                cerr = get_expression(self.context, TValidator, value=value)
            except Exception as e:
                raise Invalid(e)
            if cerr:
                raise Invalid(cerr)


@implementer(IValue)
@adapter(IEasyForm, Interface, IEasyFormForm, IField, Interface)
class FieldExtenderDefault(object):

    """ z3c.form default class for easyform fields """

    def __init__(self, context, request, view, field, widget):
        self.context = context
        self.request = request
        self.view = view
        self.field = field
        self.widget = widget

    def get(self):
        """ get default value of field from TDefault """
        fdefault = self.field.default
        efield = IFieldExtender(self.field)
        TDefault = getattr(efield, 'TDefault', None)
        return get_expression(self.context, TDefault) if TDefault else fdefault


@implementer(IFromUnicode, ILabel)
class Label(Field):

    """A Label field
    """

    def validate(self, value):
        pass

    def fromUnicode(self, str):
        """
        """
        return


@implementer(IRichLabel)
class RichLabel(Label):

    """A Rich Label field
    """
    rich_label = u''

    def __init__(self, rich_label=u'', **kw):
        self.rich_label = rich_label
        super(RichLabel, self).__init__(**kw)


LabelFactory = FieldFactory(Label, _(u'label_label_field', default=u'Label'))
RichLabelFactory = FieldFactory(
    RichLabel,
    _(u'label_richlabel_field', default=u'Rich Label')
)

LabelHandler = BaseHandler(Label)
RichLabelHandler = BaseHandler(RichLabel)


@implementer(IReCaptcha)
class ReCaptcha(TextLine):

    """A ReCaptcha field
    """


ReCaptchaFactory = FieldFactory(
    ReCaptcha,
    _(u'label_recaptcha_field', default=u'ReCaptcha')
)
ReCaptchaHandler = BaseHandler(ReCaptcha)



# -*- coding: utf-8 -*-
#
# Integeration tests specific to the mailer
#

from collective.easyform.browser.view import EasyFormForm
from collective.easyform.tests import base
from os.path import dirname
from os.path import join
from plone import api
from plone.namedfile.file import NamedFile


class TestFunctions(base.EasyFormTestCase):
    """ Test mailer action """

    def afterSetUp(self):
        super(TestFunctions, self).afterSetUp()
        ff1_id = self.folder.invokeFactory('EasyForm', 'ff1')
        self.ff1 = self.folder[ff1_id]

    def LoadRequestForm(self, **kwargs):
        request = self.layer['request']
        request.form.clear()
        prefix = 'form.widgets.'
        for key in kwargs.keys():
            request.form[prefix + key] = kwargs[key]
        return request

    def test_thankspage(self):
        """ Test thankspage """

        data = {
            'topic': 'test subject',
            'comments': 'test comments',
            'replyto': 'foo@bar.com'
        }
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertIn('Thanks for your input.', form)
        self.assertIn(
            '<span id="form-widgets-replyto" class="text-widget required textline-field">foo@bar.com</span>',  # noqa
            form)

    def test_thankspage_filter(self):
        """ Test thankspage """
        self.ff1.showAll = False
        self.ff1.showFields = ['comments']
        data = {
            'topic': 'test subject',
            'comments': 'test comments',
            'replyto': 'foo@bar.com'
        }
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertIn('Thanks for your input.', form)
        self.assertNotIn('subject', form)
        self.assertIn('test comments', form)

    def test_thankspage_radio(self):
        self.ff1.showAll = True
        field_template = api.content.create(
            self.layer['portal'], 'File', id='easyform_default_fields.xml')
        with open(join(dirname(__file__), 'fixtures', 'radio_form.xml')) as f:
            filecontent = NamedFile(f.read(), contentType='application/xml')
        field_template.file = filecontent

        data = {'yes': 'true', 'yes-empty-marker': 1}
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertIn(
            '<span id="form-widgets-yes" class="radio-widget bool-field"><span class="selected-option">yes</span></span>',  # noqa
            form
        )

    def test_show_hidden_on_thankspage(self):
        self.ff1.showAll = True
        field_template = api.content.create(
            self.layer['portal'], 'File', id='easyform_default_fields.xml')
        with open(join(dirname(__file__), 'fixtures', 'hidden_form.xml')) as f:
            filecontent = NamedFile(f.read(), contentType='application/xml')
        field_template.file = filecontent

        data = {'hide': 'hello hidden'}
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertIn(
            '<span id="form-widgets-hide" class="text-widget textline-field">hello hidden</span>',  # noqa
            form
        )

    def test_no_widget_on_thankspage_fieldset(self):
        self.ff1.showAll = True
        field_template = api.content.create(
            self.layer['portal'], 'File', id='easyform_default_fields.xml')
        with open(join(dirname(__file__),
                  'fixtures', 'fieldset_form.xml')) as f:
            filecontent = NamedFile(f.read(), contentType='application/xml')
        field_template.file = filecontent

        data = {'first-field': 'hello ff',
                'second-field-one': 'hello sf1',
                'second-field-two': 'hello sf2',
                'second-field-three': 'hello sf3'}
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn('<input id="form-widgets-second-field-one"', form)
        self.assertNotIn('<input id="form-widgets-second-field-two"', form)
        self.assertNotIn('<input id="form-widgets-second-field-three"', form)

    def test_not_showall_on_thankspage_fieldset(self):
        self.ff1.showAll = False
        field_template = api.content.create(
            self.layer['portal'], 'File', id='easyform_default_fields.xml')
        with open(join(dirname(__file__),
                       'fixtures', 'fieldset_form.xml')) as f:
            filecontent = NamedFile(f.read(), contentType='application/xml')
        field_template.file = filecontent

        data = {'first-field': 'hello ff',
                'second-field-one': 'hello sf1',
                'second-field-two': 'hello sf2',
                'second-field-three': 'hello sf3'}
        request = self.LoadRequestForm(**data)
        request.method = 'POST'
        form = EasyFormForm(self.ff1, request)()
        self.assertNotIn('<input id="form-widgets-second-field-one"', form)
        self.assertNotIn('<input id="form-widgets-second-field-two"', form)
        self.assertNotIn('<input id="form-widgets-second-field-three"', form)



# -*- coding: utf-8 -*-
from collective.easyform.tests.base import FUNCTIONAL_TESTING
from plone import api
from plone.testing import layered
from plone.testing.z2 import Browser

import doctest
import transaction
import unittest


optionflags = (
    doctest.REPORT_ONLY_FIRST_FAILURE |
    doctest.NORMALIZE_WHITESPACE |
    doctest.ELLIPSIS |
    doctest.REPORTING_FLAGS)

testfiles = (
    'browser.rst',
    'attachment.rst',
    'ssl.txt',
    'serverside_field.txt',
    '../api.py',
    '../README.txt',
)


def get_browser(layer):
    api.user.create(
        username='adm', password='secret', email='a@example.org',
        roles=('Manager', )
    )
    transaction.commit()
    browser = Browser(layer['app'])
    browser.addHeader('Authorization', 'Basic adm:secret')
    return browser


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests([
        layered(
            doctest.DocFileSuite(
                f,
                optionflags=optionflags,
                globs={'get_browser': get_browser, }
            ),
            layer=FUNCTIONAL_TESTING
        )
        for f in testfiles])
    return suite



# -*- coding: utf-8 -*-
from .validators import isTALES
from collective.easyform import easyformMessageFactory as _  # NOQA
from collective.easyform import config
from plone import api
from plone.app.textfield import RichText
from plone.autoform import directives
from plone.supermodel.model import fieldset
from plone.supermodel.model import Schema
from Products.CMFPlone.utils import safe_unicode
from zope.i18n import translate
from zope.interface import Interface
from zope.interface import provider

import zope.i18nmessageid
import zope.interface
import zope.schema.interfaces


PMF = zope.i18nmessageid.MessageFactory('plone')


@provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_submitLabel(context):
    return translate(_(u'default_submitLabel', u'Submit'))


@provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_resetLabel(context):
    return translate(_(u'default_resetLabel', u'Reset'))


@provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_thankstitle(context):
    return translate(_(u'default_thankstitle', u'Thank You'))


@provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_thanksdescription(context):
    return translate(
        _(u'default_thanksdescription', u'Thanks for your input.'))


@zope.interface.provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_actions(context):
    """ Default mail body for mailer action

        Acquire 'mail_body_default.pt' or return hard coded default
    """
    portal = api.portal.get()
    default_actions = portal.restrictedTraverse(
        'easyform_default_actions.xml', default=None)
    if default_actions:
        return safe_unicode(default_actions.file.data)
    else:
        return config.ACTIONS_DEFAULT


@zope.interface.provider(zope.schema.interfaces.IContextAwareDefaultFactory)
def default_fields(context):
    """ Default form fields

        Acquire 'easyform_default_fields.xml' or return hard coded default
    """
    portal = api.portal.get()
    default_fields = portal.restrictedTraverse(
        'easyform_default_fields.xml', default=None)
    if default_fields:
        return safe_unicode(default_fields.file.data)
    else:
        return config.FIELDS_DEFAULT


class IEasyForm(Schema):
    """Forms for Plone"""

    directives.omitted('fields_model', 'actions_model')
    fields_model = zope.schema.Text(
        title=_(u'Fields Model'),
        defaultFactory=default_fields,
    )
    actions_model = zope.schema.Text(
        title=_(u'Actions Model'),
        defaultFactory=default_actions,
    )
    submitLabel = zope.schema.TextLine(
        title=_(u'label_submitlabel_text', default=u'Submit Button Label'),
        description=_(u'help_submitlabel_text', default=u''),
        defaultFactory=default_submitLabel,
        required=False,
    )
    useCancelButton = zope.schema.Bool(
        title=_(u'label_showcancel_text', default=u'Show Reset Button'),
        description=_(u'help_showcancel_text', default=u''),
        default=False,
        required=False,
    )
    resetLabel = zope.schema.TextLine(
        title=_(u'label_reset_button', default=u'Reset Button Label'),
        description=_(u'help_reset_button', default=u''),
        defaultFactory=default_resetLabel,
        required=False,
    )
    method = zope.schema.Choice(
        title=_(u'label_method', default=u'Form method'),
        description=_(u'help_method', default=u''),
        default=u'post',
        required=False,
        vocabulary='easyform.FormMethods',
    )
    form_tabbing = zope.schema.Bool(
        title=_(u'label_form_tabbing',
                default=u'Turn fieldsets to tabs'),
        description=_(u'help_form_tabbing', default=u''),
        default=True,
        required=False,
    )
    default_fieldset_label = zope.schema.TextLine(
        title=_(u'label_default_fieldset_label_text',
                default=u'Custom Default Fieldset Label'),
        description=_(
            u'help_default_fieldset_label_text',
            default=u'This field allows you to change default fieldset label.'
        ),
        required=False,
        default=u'',
    )
    unload_protection = zope.schema.Bool(
        title=_(u'label_unload_protection',
                default=u'Unload protection'),
        description=_(u'help_unload_protection', default=u''),
        default=True,
        required=False,
    )
    CSRFProtection = zope.schema.Bool(
        title=_(u'label_csrf', default=u'CSRF Protection'),
        description=_(u'help_csrf', default=u'Check this to employ Cross-Site '
                      u'Request Forgery protection. Note that only HTTP Post '
                      u'actions will be allowed.'),
        default=True,
        required=False,
    )
    directives.write_permission(forceSSL=config.EDIT_ADVANCED_PERMISSION)
    forceSSL = zope.schema.Bool(
        title=_(u'label_force_ssl', default=u'Force SSL connection'),
        description=_(
            u'help_force_ssl',
            default=u'Check this to make the form redirect to an SSL-enabled '
                    u'version of itself (https://) if accessed via a non-SSL '
                    u'URL (http://).  In order to function properly, '
                    u'this requires a web server that has been configured to '
                    u'handle the HTTPS protocol on port 443 and forward it to '
                    u'Zope.'
        ),
        default=False,
        required=False,
    )
    formPrologue = RichText(
        title=_(u'label_prologue_text', default=u'Form Prologue'),
        description=_(
            u'help_prologue_text',
            default=u'This text will be displayed above the form fields.'
        ),
        required=False,
    )
    formEpilogue = RichText(
        title=_(u'label_epilogue_text', default=u'Form Epilogue'),
        description=_(
            u'help_epilogue_text',
            default=u'The text will be displayed after the form fields.'
        ),
        required=False,
    )
    fieldset(
        u'overrides',
        label=_('Overrides'),
        fields=[
            'thanksPageOverrideAction',
            'thanksPageOverride',
            'formActionOverride',
            'onDisplayOverride',
            'afterValidationOverride',
            'headerInjection',
            'submitLabelOverride'
        ]
    )
    directives.write_permission(
        thanksPageOverrideAction=config.EDIT_TALES_PERMISSION)
    thanksPageOverrideAction = zope.schema.Choice(
        title=_(u'label_thankspageoverrideaction_text',
                default=u'Custom Success Action Type'),
        description=_(
            u'help_thankspageoverrideaction_text',
            default=u'Use this field in place of a thanks-page designation '
                    u'to determine final action after calling '
                    u'your action adapter (if you have one). You would '
                    u'usually use this for a custom success template or '
                    u'script. Leave empty if unneeded. Otherwise, specify as '
                    u'you would a CMFFormController action type and argument, '
                    u'complete with type of action to execute '
                    u'(e.g., "redirect_to" or "traverse_to") '
                    u'and a TALES expression. For example, '
                    u'"Redirect to" and "string:thanks-page" would redirect '
                    u'to "thanks-page".'),
        default=u'redirect_to',
        required=False,
        vocabulary='easyform.CustomActions',
    )
    directives.write_permission(
        thanksPageOverride=config.EDIT_TALES_PERMISSION)
    thanksPageOverride = zope.schema.TextLine(
        title=_(u'label_thankspageoverride_text',
                default=u'Custom Success Action'),
        description=_(
            u'help_thankspageoverride_text',
            default=u'Use this field in place of a thanks-page designation '
                    u'to determine final action after calling your action '
                    u'adapter (if you have one). You would usually use '
                    u'this for a custom success template or script. '
                    u'Leave empty if unneeded. Otherwise, specify as you '
                    u'would a CMFFormController action type and argument, '
                    u'complete with type of action to execute '
                    u'(e.g., "redirect_to" or "traverse_to") '
                    u'and a TALES expression. For example, '
                    u'"Redirect to" and "string:thanks-page" would redirect '
                    u'to "thanks-page".'),
        default=u'',
        constraint=isTALES,
        required=False,
    )
    directives.write_permission(
        formActionOverride=config.EDIT_TALES_PERMISSION)
    formActionOverride = zope.schema.TextLine(
        title=_(u'label_formactionoverride_text',
                default=u'Custom Form Action'),
        description=_(
            u'help_formactionoverride_text',
            default=u'Use this field to override the form action attribute. '
                    u'Specify a URL to which the form will post. '
                    u'This will bypass form validation, success action '
                    u'adapter and thanks page.'
        ),
        default=u'',
        required=False,
        constraint=isTALES,
    )
    directives.write_permission(onDisplayOverride=config.EDIT_TALES_PERMISSION)
    onDisplayOverride = zope.schema.TextLine(
        title=_(u'label_OnDisplayOverride_text', default=u'Form Setup Script'),
        description=_(
            u'help_OnDisplayOverride_text',
            default=u'A TALES expression that will be called when the form is '
                    u'displayed. Leave empty if unneeded. The most common '
                    u'use of this field is to call a python script that '
                    u'sets defaults for multiple fields by pre-populating '
                    u'request.form. '
                    u'Any value returned by the expression is ignored. '
                    u'PLEASE NOTE: errors in the evaluation of this '
                    u'expression will cause an error on form display.'
        ),
        constraint=isTALES,
        required=False,
        default=u'',
    )
    directives.write_permission(
        afterValidationOverride=config.EDIT_TALES_PERMISSION)
    afterValidationOverride = zope.schema.TextLine(
        title=_(u'label_AfterValidationOverride_text',
                default=u'After Validation Script'),
        description=_(
            u'help_AfterValidationOverride_text',
            default=u'A TALES expression that will be called after the form is'
                    u'successfully validated, but before calling an action '
                    u'adapter (if any) or displaying a thanks page.'
                    u'Form input will be in the request.form dictionary.'
                    u'Leave empty if unneeded. The most '
                    u'common use of this field is to call a python script'
                    u'to clean up form input or to script an alternative '
                    u'action. Any value returned by the expression is ignored.'
                    u'PLEASE NOTE: errors in the evaluation of this '
                    u'expression willcause an error on form display.'
        ),
        constraint=isTALES,
        required=False,
        default=u'',
    )
    directives.write_permission(headerInjection=config.EDIT_TALES_PERMISSION)
    headerInjection = zope.schema.TextLine(
        title=_(u'label_headerInjection_text', default=u'Header Injection'),
        description=_(
            u'help_headerInjection_text',
            default=u'This override field allows you to insert content into '
                    u'the xhtml head. The typical use is to add custom CSS '
                    u'or JavaScript. Specify a TALES expression returning a '
                    u'string. The string will be inserted with no '
                    u'interpretation. PLEASE NOTE: errors in the evaluation '
                    u'of this expression will cause an error on form display.'
        ),
        constraint=isTALES,
        required=False,
        default=u'',
    )
    directives.write_permission(
        submitLabelOverride=config.EDIT_TALES_PERMISSION)
    submitLabelOverride = zope.schema.TextLine(
        title=_(u'label_submitlabeloverride_text',
                default=u'Custom Submit Button Label'),
        description=_(
            u'help_submitlabeloverride_text',
            default=u'This override field allows you to change submit button '
                    u'label. The typical use is to set label with request '
                    u'parameters. Specify a TALES expression returning a '
                    u'string. PLEASE NOTE: errors in the evaluation of this '
                    u'expression will cause an error on form display.'
        ),
        constraint=isTALES,
        required=False,
        default=u'',
    )
    fieldset(
        u'thankyou',
        label=_('Thanks Page'),
        fields=[
            'thankstitle',
            'thanksdescription',
            'showAll',
            'showFields',
            'includeEmpties',
            'thanksPrologue',
            'thanksEpilogue'
        ]
    )
    thankstitle = zope.schema.TextLine(
        title=_(u'label_thankstitle', default=u'Thanks title'),
        defaultFactory=default_thankstitle,
        required=True
    )
    thanksdescription = zope.schema.Text(
        title=_(u'label_thanksdescription', default=u'Thanks summary'),
        description=_(
            u'help_thanksdescription',
            default=u'Used in thanks page.'
        ),
        defaultFactory=default_thanksdescription,
        required=False,
        missing_value=u'',
    )
    showAll = zope.schema.Bool(
        title=_(u'label_showallfields_text', default=u'Show All Fields'),
        description=_(u'help_showallfields_text', default=u''
                      u'Check this to display input for all fields '
                      u'(except label and file fields). If you check '
                      u'this, the choices in the pick box below '
                      u'will be ignored.'),
        default=True,
        required=False,
    )
    showFields = zope.schema.List(
        title=_(u'label_showfields_text', default=u'Show Responses'),
        description=_(
            u'help_showfields_text',
            default=u'Pick the fields whose inputs you\'d like to display on '
                    u'the success page.'),
        unique=True,
        required=False,
        value_type=zope.schema.Choice(vocabulary='easyform.Fields'),  # noqa
    )
    includeEmpties = zope.schema.Bool(
        title=_(u'label_includeEmpties_text', default=u'Include Empties'),
        description=_(u'help_includeEmpties_text', default=u''
                      u'Check this to display field titles '
                      u'for fields that received no input. Uncheck '
                      u'to leave fields with no input off the list.'),
        default=True,
        required=False,
    )
    thanksPrologue = RichText(
        title=_(u'label_thanksprologue_text', default=u'Thanks Prologue'),
        description=_(
            u'help_thanksprologue_text',
            default=u'This text will be displayed above the selected field '
                    u'inputs.'
        ),
        required=False,
    )
    thanksEpilogue = RichText(
        title=_(u'label_thanksepilogue_text', default=u'Thanks Epilogue'),
        description=_(
            u'help_thanksepilogue_text',
            default=u'The text will be displayed after the field inputs.'
        ),
        required=False,
    )


class IEasyFormImportFormSchema(Interface):

    """Schema for easyform import form.
    """
    upload = zope.schema.Bytes(
        title=PMF(u'Upload'),
        required=True)


class IEasyFormThanksPage(Interface):
    """ Marker interface for thanks page """
